#pragma once
// BOUT++ header shim
#warning Header "boundary_standard.hxx" has moved to "bout/boundary_standard.hxx". Run `bin/bout-v5-header-upgrader.py` to fix
#include "bout/boundary_standard.hxx"
