#ifndef __SOLVER_FACTORY_H__
#define __SOLVER_FACTORY_H__

#ifndef _MSC_VER
#warning("Deprecated header: use #include <bout/solver.hxx> instead")
#else
#pragma message("Warning: deprecated header: use #include <bout/solver.hxx> instead")
#endif

#include <bout/solver.hxx>

#endif // __SOLVER_FACTORY_H__
